#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <time.h>
#include <datatype.h>
#include <protos.h>

/*-------------------------------*/
int main() {
/*-------------------------------*/
/*   driver program of GPU-sol   */
/*-------------------------------*/
  char finput[] = "input";
  options_t *opts;
  matrix_t *mat;
  precon_t *prec;
  coo_t *h_coo;
  int i,n,nnz,*perm1,*perm2,*perm3;
  double *h_b,*h_x,*sol,*diag;
  double *d_b,*d_x;
/*----------------------- init */
  if (gpusol_init()) {
   printf("GPU-Solv Init Error\n");
   return 1;
  }
  h_b=h_x=sol=diag=d_b=d_x=NULL;
  perm1 = perm2 = NULL;
/*------------ read input file */
  Calloc(opts, 1, options_t);
  read_input(finput, opts);
/*----------- first step: ----------*/ 
/*-- alloc 'mat' & 'prec' structure */
/*----------------------------------*/
  Calloc(mat,  1, matrix_t);
  Calloc(prec, 1, precon_t);
/* load matrix,stored in COO struct */
  Calloc(h_coo, 1, coo_t);
  read_coo_MM(h_coo, opts);
  n = h_coo->n;  nnz = h_coo->nnz;
/*-------- convert to CSR in 'mat' */
  mat->n = n;  mat->nnz = nnz;
  Calloc(mat->h_csr, 1, csr_t);
  COO2CSR(h_coo, mat->h_csr);
/*-------- exact sol */
  Malloc(sol, n, double);
  for (i=0; i<n; i++) sol[i] = 1.0;
/*-------- artificial rhs=A*sol */
  Malloc(h_b, n, double);
  spmv_csr_cpu(mat->h_csr, sol, h_b);
/*-------- diag scaling */
  diag_scal(mat->h_csr, h_b, opts, &diag);
/*------ get reordering (MMD/ND/RCM) */
  reorder_sym(mat->h_csr, opts, &perm1);
/*------ permute mat */
  perm_mat_sym(mat->h_csr, perm1);
/*------ mc reordering */
  reorder_mc(mat->h_csr, prec, opts, &perm2);
/*------ permute mat */
  perm_mat_sym(mat->h_csr, perm2);
/*------ partitioning & reordering */
/*------ only for block-iluk prec */
  reorder_dd(mat->h_csr, prec, opts, &perm3);
/*------ permute mat */
  perm_mat_sym(mat->h_csr, perm3);
/*------ permute rhs */
  perm_vec(n, h_b, perm1);
  perm_vec(n, h_b, perm2);
  perm_vec(n, h_b, perm3);
/*------ setup mat & copy to device */
  setup_matrix(mat, opts);
/*------ setpu precon */
  setup_precon(mat, prec, opts);
/*------ copy rhs to device */
  d_b = (double*)cuda_malloc(n*sizeof(double));
  memcpyh2d(d_b, h_b, n*sizeof(double));
/*------ init guess */
  Malloc(h_x, n, double);
  for (i=0; i<n; i++) h_x[i] = 0.0;
/*------ copy init guess to device */
  d_x = (double*)cuda_malloc(n*sizeof(double));
  memcpyh2d(d_x, h_x, n*sizeof(double));
/*------ preconed itersol */
/*------ GMRES/CG ------- */
  pitsol(mat, prec, opts, d_x, d_b);
/*------ copy result to host */
  memcpyd2h(h_x, d_x, n*sizeof(double));
/*------ check residue norm */
  resd_norm(mat->h_csr, h_x, h_b, 
  &(opts->result));
/*------ permute solution */
  iperm_vec(n, h_x, perm3);
  iperm_vec(n, h_x, perm2);
  iperm_vec(n, h_x, perm1);
/*------ scale solution */
  scal_vec(n,diag, h_x);
/*------ check solution */
  err_norm(n, h_x, sol, &(opts->result));
/*------ output result */
  output_result(mat, opts);
/*--------------------- *
  for (i=0; i<min(n,10);i++)
    printf(" %f\n", h_x[i]);
*/
/*------ done, free mem */
  free_opts(opts);
  free_matrix(mat);
  free_precon(prec);
  free_coo(h_coo);
  Free(h_b); cuda_free(d_b);
  Free(h_x); cuda_free(d_x);
  Free(sol); Free(diag);
  Free(perm1); Free(perm2); Free(perm3);
/*------finalize & check error */
  gpusol_finalize();
}

