#include "spmv.h"

/*-----------------------------------------*/
int main (int argc, char **argv) {
/*-----------------------------------------*
 *           y = A * x
 *     CPU CSR        kernel
 *     GPU CSR-Scalar kernel
 *     GPU CSR-Vector kernel
 *     GPU JAD        kernel
 *     GPU DIA        kernel
 *-----------------------------------------*/
  int i, n, ret;
  REAL *x, *y0, *y;
  double e1, e2, e3, e4;
  struct coo_t coo;
  struct csr_t csr;
  struct jad_t jad;
  struct dia_t dia;
/*-----------------------------------------*/
  if (argc != 2) {
    fprintf(stdout,"Usage: spmv mtxfile\n");
    return(1);
  }
  srand (SEED);
/*------------ Init GPU */
  cuda_init(argc, argv);
/*------------ output header */
  print_header();
/*---------- Read from Martrix Market file */
  read_coo_MM(&coo, argv[1]);
/*-----------------------------------------*/
  n = coo.n;
  x  = (REAL *) malloc(n*sizeof(REAL));
  y0 = (REAL *) malloc(n*sizeof(REAL));
  y  = (REAL *) malloc(n*sizeof(REAL));
/*---------- randomly init x */
  for (i=0; i<n; i++)
    x[i] = rand() / (RAND_MAX + 1.0);
/*---------- convert COO to CSR */
  COO2CSR(&coo, &csr);
/*---------- CPU SpMV CSR kernel */
  spmv_csr_cpu(&csr, x, y0);
/*---------- GPU SpMV CSR-scalar kernel */
  spmv_csr_scalar(&csr, x, y);
  e1 = error_norm(y0, y, n);
  fprintf(stdout, "err norm %.2e\n", e1);
/*---------- GPU SpMV CSR-vector kernel */
  spmv_csr_vector(&csr, x, y);
  e2 = error_norm(y0, y, n);
  fprintf(stdout, "err norm %.2e\n", e2);
/*---------- convert CSR to JAD */
  CSR2JAD(&csr, &jad);
/*---------- GPU SpMV JAD kernel */
  spmv_jad(&jad, x, y);
  e3 = error_norm(y0, y, n);
  fprintf(stdout, "err norm %.2e\n", e3);
/*---------- convert CSR to DIA */
  if (CSR2DIA(&csr, &dia)) {
/*---------- GPU SpMV DIA kernel */
    spmv_dia(&dia, x, y);
    e4 = error_norm(y0, y, n);
    fprintf(stdout, "err norm %.2e\n", e4);
  } else {
    printf("\n=== [GPU] DIA-vector Kernel ===\n");
    printf("  exceed max diag allowed\n");
  }
/*---------- check error */
  cuda_check_err();
/*---------- Done, free */  
  FreeCOO(&coo);  FreeCSR(&csr);
  FreeJAD(&jad);  FreeDIA(&dia);
  free(x);   free(y0);  free(y);
}

